/* N-body calculation using Greenspan 1st-order discrete mechanics.	*/
/* Uses N*N processors, communication by spreading and reduction,	*/
/* explict loops for xyz.  All processors store particle information	*/
/* and calculate velocities, positions.					*/
/* Uses dynamic shape allocation to permit more flexible use.		*/
/* Shape has extra dimension to permit useful tests for small N.	*/
/*******    Provides for checkpointing using the Data Vault	  *******/
/*******    timed with system wall-clock timer.		          *******/
/* File:  nb1dsc.cs							*/

#include <stdio.h>
#include <fcntl.h>
#include <sys/time.h>
#include <math.h>
#include <stdlib.h>
#include <cscomm.h>
#include <cm/cmfs.h>
#include <cm/cm_file.h>

/* Max number of bodies in the calculation.  N = 128 => 16k vprocs */

int N=128;	/* Max number of bodies permitted		*/
int   NB;	/* Number of bodies in system, <= N		*/
int   PX;	/* Controls how much state is printed		*/
float G;	/* Gravitational constant			*/
float MINF;	/* Minimum CM single-float			*/

#define TWOPI      6.2831853f

/* Particles, and pairs of particles.  All entries have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [][]pair;		/* Will be [N][N]	*/

/* All parallel variables are declared as pointers so 'pair' can be	*/
/* dynamic.  State variables will be allocated as x0[3], etc.;		*/
/* w, m, mj, r0, are floats						*/

/* x0, x1 - current, new positions;	v0, v1 - current, new velocities */
/* m - mass;		mj - scaled mass;	ac - acceleration	 */
/* w, ax - working	r0 - distance (X0)	d0 - displacement (X0)	 */

float:pair *x0, *x1, *v0, *v1, *m, *mj, *ac, *d0, *r0, *ax, *w;

/* i - row number;    j - column number;	*/

int:pair *i, *j;

/* sm   - processor is part of computation = diag | act	*/
/* diag - processor is on diagonal			*/
/* act  - processor is active pair (off-diag)		*/

bool:pair *sm, *diag, *act;

/* Allocate global parallel variables, pair must have been allocated	*/
/* State for checkpointing is allocated in one block  *******************/

int Chksz;	/* Size of checkpoint block (bools) */

void create_globals()
 { int fsz, isz, bsz, vsz;
   fsz = boolsizeof(float:pair);  vsz = 3*fsz;  Chksz = 8*fsz;
   isz = boolsizeof(int:pair);
   bsz = boolsizeof(bool:pair);
   /* Checkpoint state */
   x0 = palloc(pair, Chksz);
   v0 = &x0[3];  m = &x0[6];  mj = &x0[7];
   /* Other float vectors[3] */
   x1 = palloc(pair, vsz);  v1 = palloc(pair, vsz);
   ac = palloc(pair, vsz);  ax = palloc(pair, vsz);
   d0 = palloc(pair, vsz);
   /* Other floats */
   r0 = palloc(pair, fsz);  w  = palloc(pair, fsz);
   /* Ints and bools */
   i  = palloc(pair, isz);  j = palloc(pair, isz);
   sm = palloc(pair, bsz);  diag = palloc(pair, bsz);  act = palloc(pair, bsz);
 }

/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { /* Initialize coordinates */
      *i = pcoord(0);  *j = pcoord(1);
      /* Identify active processors */
      *sm = (*i < NB) & (*j < NB);
      *act = *sm & (*i != *j);  *diag = *sm & (*i == *j);
    }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */
/* Leaves data for Pi in processors [c][i][0].				*/

void init()
 { int k;  float a, b, c;
   with (pair)
   {for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       [k][0]x0[0] = a;  [k][0]x0[1] = b;  [k][0]x0[2] = c;
     };
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       [k][0]v0[0] = a;  [k][0]v0[1] = b;  [k][0]v0[2] = c;
     };
    for ( k = 0; k < NB; k++ )
      { fscanf(datap, "%e", &a);  [k][0]*m = a; };
    fscanf(datap, "%e", &G);
    /* Spread data across rows */
    x0[0] = copy_spread(&x0[0], 1, 0);
    x0[1] = copy_spread(&x0[1], 1, 0);
    x0[2] = copy_spread(&x0[2], 1, 0);
    v0[0] = copy_spread(&v0[0], 1, 0);
    v0[1] = copy_spread(&v0[1], 1, 0);
    v0[2] = copy_spread(&v0[2], 1, 0);
    *m  = copy_spread(m,  1, 0);
   }
 }

/* Insert nx asteroids at radius ar with total mass am.  This assumes	*/
/* that an orbit of radius 1 about origin corresponds to velocity 2*pi	*/
/* (e.g., au-earth-yr units).  Uses parallel trig.			*/

void mkast(int nx, float ar, float am)
 { float:pair sx, cx, arg;
   float spd;  int nn;
   nn = NB + nx;
   spd = TWOPI/sqrt(ar);
   printf("Making %d asteroids, r = %9.5f, spd = %9.5f, tot mass = %9.2e\n",
		  nx,		     ar,	  spd,		    am);
   with (pair) 
   { *sm = (*i < nn) & (*j < nn);
     *act = *sm & (*i != *j);  *diag = *sm & (*i == *j);
     where ( (NB <= *i) & (*i < nn) & *sm )
      { arg = (TWOPI*(*i - NB))/nx;
        sx = sin(arg);  cx = cos(arg);
        x0[0] =  cx*ar;  v0[0] = (-sx)*spd;
        x0[1] =  sx*ar;  v0[1] =  cx*spd;
        x0[2] = 0.0;     v0[2] = 0.0;
        *m = am/nx;
      }}  /* end where, end with (pair) */
   NB = nn;
 }


/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.					*/

void print_state()
 { int k, np;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < np; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, [k][k]x0[0], [k][k]x0[1], [k][k]x0[2],
               [k][k]v0[0], [k][k]v0[1], [k][k]v0[2], [k][k]*m);
 }

/* Print x-components of xx in nx by nx processors (for testing). */

/******************************************************************
void print_x(char lbl[], float:current *xx, int nx)
 { int u, v;
   printf("\n%s\n", lbl); 
   for ( u = 0; u < nx; u++)
    { for ( v = 0;  v < nx; v++)
        printf("%9.2e ", [u][v]*xx);
      printf("\n");
    }
 }
*******************************************************************/

/* Calculate accelerations.  r0, d0 already give displacement, distance  */
/* for current position.  Called with all processors active.		 */

void accel()
 { float:pair d1[3], r1;  int u;
  everywhere
  {*w = 0.0f;			/* To accumulate sum of squares (later) */
   for ( u = 0;  u < 3;  u++ )
     { ac[u] = 0.0f;  ax[u] = MINF;
       where (*diag) ax[u] = x1[u];	/* Prepare to spread */
       d1[u] = spread(ax[u], 0, CMC_combiner_max);
     }
   /* print_x("accel, d1 after spread 1", d1, NB);	 For testing */
   where (*act)
    { /* Now d1 = [j][j]x1,  x1 = [i][i]x1 */
      for ( u = 0;  u < 3;  u++ )
       { d1[u] = d1[u] - x1[u];		/* New displacement component */
	 *w += d1[u]*d1[u];		/* For sum of squares */
       }
      r1 = sqrt(*w);		/* Euclidean distance */
      /* print_x("accel, r1", &r1, NB);	For testing */
      *w = *mj/(*r0*r1*(*r0+r1));	/* Scalar multiplier */
    } /* end where (*act) */
   for ( u = 0;  u < 3;  u++ )
     { ac[u] = *w*(d0[u]+d1[u]);	/* Acceleration component */
       ac[u] = spread(ac[u], 1, CMC_combiner_add);  /* Sum accelerations */
     }
   /* print_x("accel, final ac", ac, NB);	For testing */
  }  /* end everywhere */
 }

/* Calculate initial accelerations.  Called with all proc's active. */

void accel1()
 { int u;
   everywhere
  {*w = 0.0f;			/* To accumulate sum of squares (later) */
   for ( u = 0;  u < 3;  u++ )
     { ac[u] = 0.0f;  ax[u] = MINF;
       where (*diag) ax[u] = x0[u];	/* Prepare to spread */
       d0[u] = spread(ax[u], 0, CMC_combiner_max);
     }
   /* print_x("accel1, d0 after spread 1", d0, NB);	For testing */
   where (*act)
    { /* Now d0 = [j][j]x0,  x0 = [i][i]x0 */
      for ( u = 0;  u < 3;  u++ )
       { d0[u] = d0[u] - x0[u];		/* New displacement component */
	 *w += d0[u]*d0[u];		/* For sum of squares */
       }
      *r0 = sqrt(*w);		/* Euclidean distance */
      /* print_x("accel1, r0", r0, NB);	For testing */
      *w = *mj/(*r0**w);	/* Scalar multiplier */
    } /* end where (*act) */
   for ( u = 0;  u < 3;  u++ )
    { ac[u] = *w*d0[u];		/* Acceleration component */
      ac[u] = spread(ac[u], 1, CMC_combiner_add);  /* Sum accelerations */
    }
   /* print_x("accel1, final ac", ac, NB);		For testing */
  }  /* end everywhere */
 }

/* Greenspan's 1st-order method.  Performs ns steps of time dt		*/
/* with solution tolerance eps.  Returns number of evaluations.		*/
/* Leaves number of iterations which were stopped at limit in cfail.	*/
/* Some printing for debugging is provided in comments.			*/

int cfail;   /* Counts number of iterations that failed to converge */

int green1(int ns, float dt, float eps)
 { int k, ic, v, lim=10, u;  float hdt, mx;  float:pair nx, px;
   v = 0;  hdt = 0.5*dt;
   with (pair) everywhere
   { /* print_x("green1, x0 initially", x0, NB);	For testing */
     /* print_x("green1, v0 initially", v0, NB);	For testing */
     for (k = 0; k < ns; k++)
       { ic = 0;  accel1();
	 /* This is a do - while - repeat loop: */
   	 do
          { v++;  ic++;
	    px = 0.0f;	/* For max change */
	    where (*sm)
	    { for ( u = 0;  u < 3;  u++ )
	       { nx = v0[u] + dt*ac[u];			/* New velocity	*/
		 px >?= fabs(nx - v1[u]);		/* Max change	*/
		 v1[u] = nx;
		 nx = x0[u] + hdt*(v0[u] + v1[u]);	/* New position	*/
		 px >?= fabs(nx - x1[u]);		/* Max change	*/
		 x1[u] = nx; 
	       } /* end for u */
	      mx = (>?= px);				/* Global max */
   	      /* print_x("green1, v1", v1, NB);		For testing */
   	      /* print_x("green1, x1", x1, NB);		For testing */
	      } /* end where (*sm) */
	    if ( (mx < eps) || ( ic == lim ) ) break;
	    accel(); }
	 while (1);
         where (*sm)
	   for ( u = 0;  u < 3;  u++ ) { x0[u] = x1[u];  v0[u] = v1[u]; }
	 if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  with (pair) */
   return(v);
 }


/*  Elapsed time from ts to tf in msec.  */

float tv_diff(struct timeval *ts, struct timeval *tf)
 { return ( 1000*(tf->tv_sec - ts->tv_sec)
	      + (tf->tv_usec - ts->tv_usec)/1000.0 );
 }


/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	  0 - just performance data					*/
/*	< 0 - all							*/
/*	> 0 - number given						*/
/* Commands:								*/
/*	  c - continue calculation (same parameters)			*/
/*	  x - write checkpoint						*/
/*	  r - resume from checkpoint (not before initial run)		*/
/*	  q - quit							*/

int dvfd;		/* File descriptor for datavault */
FILE *chstrm;		/* File descriptor for host file */

void main()
 { int ns, np, k, vl, ibf[3], hfsz, hisz, nn;
   struct timeval ts, tf;	/* For front-end timer */
   float dt, eps, ar, am, tm, fbf[3];
   char fnm[50];
   /* Initialize the pair shape with suitable geometry */
   printf("Enter N for [N][N] shape> ");
   scanf("%d", &N);
   allocate_shape(&pair, 2, N, N);
   create_globals();
   hfsz = sizeof(float);  hisz = sizeof(int);
   /* Initialize MINF = min CM float = max over empty set */
   with (pair)
    { *sm = 0;  *x0 = 0.0;  where (*sm) MINF = (>?= *x0); }
   /* Loop for test cases */
   while ( printf("Data file (or c,x,r,q)> "),
	   scanf("%s", fnm),
	   ( fnm[0] != 'q' ) || ( fnm[1] > 0 ) )
    if ( ( fnm[0] == 'x' ) && ( fnm[1] == 0 ) )
     with (pair)
      { /* Checkpoint in user-specified files */
 	printf("Checkpoint file> ");  scanf("%s", fnm);
        CM_timer_clear(1);  CM_timer_start(1);
        dvfd = CMFS_creat(fnm, 0600);
	gettimeofday(&ts, NULL);
	CMFS_write_file_always(dvfd, x0, Chksz);
	gettimeofday(&tf, NULL);
	CMFS_close(dvfd);
	CM_timer_stop(1);  CM_timer_print(1);
	printf(" Unix wall clock elapsed time for data xfer:  %f msec.\n",
					   tv_diff(&ts, &tf));
	chstrm = fopen(fnm, "w");
	fbf[0] = dt;  fbf[1] = eps;  fbf[2] = tm;
	fwrite(fbf, hfsz, 3, chstrm);
	ibf[0] = N;   ibf[1] = ns;    ibf[2] = np;
	fwrite(ibf, hisz, 3, chstrm);  fclose(chstrm);
      }
    else if ( ( fnm[0] == 'r' ) && ( fnm[1] == 0 ) )
      { /* Resume from user-specified files - CM must have been */
	/* initialized with proper geometry by previous run.    */
	if ( printf("Checkpoint file > "),
	     scanf("%s", fnm),
	     (chstrm = fopen(fnm, "r")) == NULL )
	  printf("File not found.\n");
        else
	{ fread(fbf, hfsz, 3, chstrm);
	  dt = fbf[0];  eps = fbf[1];  tm = fbf[2];
	  fread(ibf, hisz, 3, chstrm);  fclose(chstrm);
	  nn = ibf[0];  ns = ibf[1];  np = ibf[2];
	  if (  (nn != N) )
	    printf("Wrong shape - can't resume.\n");
	  else
	   with (pair)
	   { CM_timer_clear(1);  CM_timer_start(1);
             dvfd = CMFS_open(fnm, CMFS_O_RDONLY);
	     gettimeofday(&ts, NULL);
	     CMFS_read_file_always(dvfd, x0, Chksz);
	     gettimeofday(&tf, NULL);
	     CMFS_close(dvfd);
	     CM_timer_stop(1);  CM_timer_print(1);
	     printf(" Unix wall clock elapsed time for data xfer:  %f msec.\n",
					     tv_diff(&ts, &tf));
	   }}}
  else
   { if ( ( fnm[0] != 'c' ) || ( fnm[1] != 0 ) )
    { printf("Print ctl> ");  scanf("%d", &PX);
      /* Performs np integrations, each ns steps of dt */
      datap = fopen(fnm, "r");
      fscanf(datap, "%d", &NB);
      setup();   /* Setup geometric structure */
      printf("Initialize....\n");
      init();    /* Read data for individual bodies */
      /* Read "asteroid" descriptions */
      while ( fscanf(datap, "%d", &ns), ns > 0)
       { fscanf(datap, "%e%e", &ar, &am);  mkast(ns, ar, am); }
      fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);  fclose(datap);
      tm = 0.0;
      printf("\n   NB = %d", NB);
      printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
                    G, dt, eps);
      printf("\n simulated time = 0");
      print_state();
      /* Spread masses and scale */
      with (pair) where (*act) { *m = [*j][*j]*m;  *mj = G**m; }
    } /* end initialization */
      for ( k = 0; k < np; k++ )
       { CM_timer_clear(1);  CM_timer_start(1);
         cfail = 0;
         vl = green1(ns, dt, eps);
         CM_timer_stop(1);  CM_timer_print(1);
	 tm += ns*dt;
         printf(
        "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
			      tm,   vl,            cfail);
         print_state();
       };  /* end for */
   }  /* end one run */
 }

/*********************************************************************
  Data files:  as for nb3dg2.cs
*************************************************************************/
